import shutil
import tempfile
from itertools import groupby
from logging import Logger, getLogger
from pathlib import Path
from zipfile import BadZipFile, ZipFile

import requests
from apistar.exceptions import ErrorResponse
from lxml import etree
from PIL import Image, ImageOps
from teklia_toolbox.requests import should_verify_cert

from arkindex_worker.models import Element
from arkindex_worker.worker.metadata import MetaType
from worker_file_import.page_xml import PageXmlParser
from worker_file_import.utils import hash_image

logger: Logger = getLogger(__name__)


def is_transkribus_export(datafile) -> bool:
    try:
        # Opening the file with ZipFile will raise a BadZipFile exception
        # when the file is not a valid ZIP file.
        with ZipFile(datafile["local_path"]) as archive:
            # To detect that an archive is a Transkribus export, we look for a `mets.xml` file,
            # which should be present in every export, under the subdirectories for each document.
            return any(
                filename.endswith("/mets.xml") for filename in archive.namelist()
            )
    except BadZipFile:
        logger.error(
            f"File {datafile['name']} does not appear to be a valid ZIP archive."
        )
        return False


class TranskribusElement:
    def __init__(
        self,
        worker,
        archive,
        image_filename,
        xml_filename,
        folder_element,
        element_type,
        paragraph_type,
        line_type,
    ):
        self.worker = worker
        self.archive = archive
        self.image_filename = image_filename
        self.xml_filename = xml_filename
        self.folder_element = folder_element
        self.element_type = element_type
        self.paragraph_type = paragraph_type
        self.line_type = line_type

    def save_image(self):
        logger.info(f"Extracting image {self.image_filename}")

        # Extract the image first before editing it.
        # Using Image.open(ZipFile.open()) is 15 times slower than extracting first!
        # Do not save in get_working_dir() here as it would become an artifact
        _, img_path = tempfile.mkstemp()
        with (
            self.archive.open(self.image_filename) as source,
            Path(img_path).open("wb") as destination,
        ):
            shutil.copyfileobj(source, destination)

        with Image.open(img_path) as img:
            # Convert to RGB if needed
            if img.mode != "RGB":
                img = img.convert("RGB")

            # Apply the rotation/transposition from the EXIF orientation tag if it is used
            img = ImageOps.exif_transpose(img)
            img.save(img_path, format="JPEG")

        return img_path

    def upload_image(self):
        image_path = self.save_image()
        logger.info(f"Creating image {self.image_filename}")
        created = False
        try:
            self.image = self.worker.api_client.request(
                "CreateImage", body={"hash": hash_image(image_path)}
            )
            created = True
        except ErrorResponse as e:
            if e.status_code == 400 and "id" in e.content:
                self.image = self.worker.api_client.request(
                    "RetrieveImage", id=e.content["id"]
                )
            else:
                raise
        if created:
            logger.info(f"Uploading image {self.image_filename} as {self.image['id']}")
            try:
                with Path(image_path).open("rb") as f:
                    requests.put(
                        self.image["s3_put_url"],
                        data=f,
                        verify=should_verify_cert(self.image["s3_put_url"]),
                    )
            finally:
                Path(image_path).unlink()

        logger.info(f"Validating image {self.image['id']}")
        # Store the returned image as it might include updated dimensions
        self.image = self.worker.api_client.request(
            "PartialUpdateImage", id=self.image["id"], body={"status": "checked"}
        )

    def upload_transcriptions(self):
        element = None
        logger.info(f"Importing transcript {self.xml_filename}")
        try:
            # Read and parse the xml file
            xml_text = self.archive.read(self.xml_filename)
            parser = PageXmlParser(self.worker, etree.fromstring(xml_text), self.image)
            element, _ = parser.save(
                parent_element=self.folder_element,
                image_id=self.image["id"],
                page_type=self.element_type,
                paragraph_type=self.paragraph_type,
                line_type=self.line_type,
            )

            if element is not None:
                self.worker.create_metadata(
                    Element(element),
                    type=MetaType.Text,
                    name="original_image_name",
                    value=parser.pagexml_page.page.image_name
                    or Path(self.image_filename).name,
                )
        except AssertionError as e:
            logger.error(f"Failed importing transcript {self.xml_filename}: {e}")
        except ErrorResponse as e:
            logger.error(
                f"Failed importing transcript {self.xml_filename}: {e.status_code} - {e.content}"
            )
        except requests.exceptions.RequestException as e:
            logger.error(f"Failed importing transcript {self.xml_filename}: {e}")
        return element

    def run(self):
        try:
            self.upload_image()
            return self.upload_transcriptions()
        except ErrorResponse as e:
            logger.error(
                f"Failed importing {self.image_filename} and {self.xml_filename}: {e.status_code} - {e.content}"
            )
        except Exception as e:
            logger.error(
                f"Failed importing {self.image_filename} and {self.xml_filename}: {e}"
            )
        return []


class TranskribusFolder:
    def __init__(
        self,
        worker,
        archive,
        filenames,
        title,
        project_id,
        folder_type,
        element_type,
        paragraph_type,
        line_type,
        parent_element=None,
    ):
        self.worker = worker
        self.archive = archive
        self.filenames = filenames
        self.title = title
        self.project_id = project_id
        self.parent_element = parent_element
        self.folder_type = folder_type
        self.element_type = element_type
        self.paragraph_type = paragraph_type
        self.line_type = line_type
        self.folder = None

    def get_or_create_folder(self):
        if self.parent_element is None:
            search = self.worker.api_client.paginate(
                "ListElements",
                corpus=self.project_id,
                name=self.title,
                type=self.folder_type,
            )
        else:
            search = self.worker.api_client.paginate(
                "ListElementChildren",
                id=self.parent_element["id"],
                name=self.title,
                type=self.folder_type,
                recursive=True,
            )

        for folder in search:
            # The name filter on ListElements will match if the name of any element contains self.title (case insensitive)
            # so we need to check again for strict equality.
            if folder["name"] == self.title:
                self.folder = folder
                return False

        logger.info(f"Creating {self.folder_type} {self.title}")
        if self.parent_element:
            self.folder = self.worker.create_sub_element(
                Element(self.parent_element),
                type=self.folder_type,
                name=self.title,
                slim_output=False,
            )
        else:
            self.folder = self.worker.api_client.request(
                "CreateElement",
                body={
                    "corpus": self.project_id,
                    "type": self.folder_type,
                    "name": self.title,
                    "worker_run_id": self.worker.worker_run_id,
                },
            )
        return True

    def run(self):
        elements = []
        try:
            created = self.get_or_create_folder()
            elements.append(self.folder)
            if not created:
                logger.info(
                    "Skipping existing {} {}".format(
                        self.folder_type, self.folder["name"]
                    )
                )
                return elements

            # Group files by name without extension: each group will hold the PAGE XML file and the image
            for name, filenames in groupby(
                sorted(self.filenames, key=lambda path: path.stem),
                key=lambda path: path.stem,
            ):
                # Ignore the METS file and the Transkribus metadata file
                if name in ("mets", "metadata"):
                    continue

                filenames = list(filenames)
                if len(filenames) != 2:
                    logger.warning(f"Expected 2 files for page {name}; skipping")
                    continue

                if filenames[0].suffix == ".xml":
                    xml_filename, image_filename = filenames
                else:
                    image_filename, xml_filename = filenames

                elements.append(
                    TranskribusElement(
                        worker=self.worker,
                        archive=self.archive,
                        image_filename=str(image_filename),
                        xml_filename=str(xml_filename),
                        folder_element=self.folder,
                        element_type=self.element_type,
                        paragraph_type=self.paragraph_type,
                        line_type=self.line_type,
                    ).run()
                )
        except ErrorResponse as e:
            logger.error(
                f"Failed importing {self.folder_type} {self.title}: {e.status_code} - {e.content}"
            )
        except Exception as e:
            logger.error(f"Failed importing {self.folder_type} {self.title}: {e}")
        return elements


class TranskribusImporter:
    def __init__(
        self,
        worker,
        archive_path,
        *,
        parent_element=None,
    ):
        self.archive_path = archive_path
        self.worker = worker
        self.parent_element = parent_element
        self.folder_type = self.worker.process_information["folder_type"]
        self.element_type = self.worker.process_information["element_type"]
        self.paragraph_type = "paragraph"
        self.line_type = "text_line"
        # Check if paragraph and text_line element types exist on the target project; if not create them
        self.project_id = self.worker.process_information["corpus"]
        project = self.worker.api_client.request("RetrieveCorpus", id=self.project_id)
        if not any(t["slug"] == "paragraph" for t in project["types"]):
            self.worker.api_client.request(
                "CreateElementType",
                body={
                    "slug": "paragraph",
                    "display_name": "Paragraph",
                    "corpus": self.project_id,
                },
            )
        if not any(t["slug"] == "text_line" for t in project["types"]):
            self.worker.api_client.request(
                "CreateElementType",
                body={
                    "slug": "text_line",
                    "display_name": "Text Line",
                    "corpus": self.project_id,
                },
            )

    def run(self):
        elements = []
        # Archive structure: <document ID>/<document title>/[<optional subdir>/]<page number>_<page ID>.[xml,jpg,tif…]
        # Also includes a log.txt file at the root and a mets.xml in each document folder.
        with ZipFile(self.archive_path, "r") as archive:
            paths = sorted(map(Path, archive.namelist()))

            def grouper(path):
                """
                Group all files by `docID/title`.
                Path('a/b/c/d').parents returns an iterator for the paths a/b/c, a/b, a, and `.`.
                With an absolute path, we would get `/` as the last parent instead of `.`.
                The folders that we are looking for are a/b, so we want the third-to-last item.
                We cannot access a parent by index before Python 3.10, so until then, we also use a list().
                """
                parents = list(path.parents)
                if len(parents) < 3:
                    # This is not a document directory.
                    return None
                return parents[-3]

            for parent, filenames in groupby(paths, key=grouper):
                # Ignore any file found outside of a document, such as `log.txt`.
                if not parent:
                    continue

                try:
                    elements += TranskribusFolder(
                        self.worker,
                        archive,
                        filenames,
                        # `parent` here is <doc ID>/<doc title>, so we extract the title
                        title=parent.name,
                        project_id=self.project_id,
                        parent_element=self.parent_element,
                        folder_type=self.folder_type,
                        element_type=self.element_type,
                        paragraph_type=self.paragraph_type,
                        line_type=self.line_type,
                    ).run()
                except Exception as e:
                    logger.error(f"Failed importing folder {parent}: {e}")

        return elements
